----------------
-- File globals
----------------
local shortName = "SomeTracker"
local D = LibStub("DoomCore-1.2")
local Addon = D.Addon(shortName, "Some Tracker")
Addon.version = 0.3

local K, setIndex, colUnpack, D_updateFrame = D.K, D.setIndex, D.colUnpack, D.updateFrame
local assertType, nilSort, roundString = D.assertType, D.nilSort, D.roundString
local D_NodeCrawler = D.NodeCrawler

local templateCache = D_NodeCrawler:New()
local player = {}
local inWorld
local unique = D_NodeCrawler:New()
local log = {}
local logKeys = {}
local logTotals = D_NodeCrawler:New()

local frames = {}
local spacing = 5
local fadeDuration = 0.6
--local missingIcon = "Interface\\Icons\\INV_Misc_QuestionMark"

--------
-- Core
--------

function Addon:OnInitialize()
  self:Register{
    media = "LibSharedMedia-3.0",
    masque = "Masque",
    templates = "SomeTemplates-0.5"
  }

  self:TrackEvent("COMBAT_LOG_EVENT_UNFILTERED")
  self:TrackEvent("PLAYER_ENTERING_WORLD")
  self:TrackMultiEvent("UpdateSpec", "PLAYER_SPECIALIZATION_CHANGED", "LEARNED_SPELL_IN_TAB")
  self:TrackEvent("UpdateInv","BAG_UPDATE")
end

function Addon:OnLoad(registered)
  self:UpdateSpec()
  self:UpdateInv()
  self.core.hiscore = self.core.hiscore or {}

  for trackerName, tracker in self:GetAll(self.core, "tracker") do
    self:BuildTracker(trackerName)
  end
end

function Addon:Migrate()
  local oldVersion = self:GetVersion() or 0
  self:SetVersion(Addon.version)
  if oldVersion >= 0.3 then return end
  self.core.hiscore = {}
  for trackerName, kind, meterName, meter in self:GetAll(self.core, "meter", 4) do
    meter.kind = kind
  end
  if oldVersion >= 0.2 then return end
  for trackerName, tracker in self:GetAll(self.core, "tracker") do
    self:Assoc(tracker)
  end
  if oldVersion > 0 then return end
  self:MigrateDB({}, { "Default Damage" }, { "Default Healing" })
  self:Make(self.core, "tracker", "Default Damage")
  self:Make(self.core, "tracker", "Default Healing")
  return true
end


function Addon:OnReset(registered)
  if registered == false and self:Migrate() then return end
  local damage = self:Make(self.core, "tracker", "Default Damage")
  if not rawget(damage,"overflow") then damage.overflow = "addition" end
  local healing = self:Make(self.core, "tracker", "Default Healing")
  healing.trackerType = "healing"
  if not rawget(healing, "offsetX") then healing.offsetX = 30*abs(damage.offsetX)/damage.offsetX - damage.offsetX end
end


-------------------
-- Tracker editing
-------------------

function Addon:Assoc(tracker)
  local assoc = D_NodeCrawler:New()
  for kind, name, meter in self:GetAll(tracker, "meter", 3) do
    assoc:Set({ kind, name, name }, meter)
    for alsoName, also in self:GetAll(meter, "also") do
      assoc:Set({ kind, alsoName, name }, meter)
    end
    for subKind,subName, subMeter in self:GetAll(meter, "subMeter", 3) do
      assoc:Set({ subKind, subName, name }, meter)
    end
  end
  tracker.assoc = assoc
end

function Addon:AddTracker(info, name)
  assertType(info,"table", name,"string")
  local tracker = self:Make(self.core, "tracker", name)
  if not tracker then return end
  self:BuildTracker(name)
end
function Addon:RenameTracker(info, new)
  assertType(info,"table", new,"string")
  local core = self.core
  if core[new] and core[new].type then return end
  local old = info[1]
  core[new] = core[old]
  core[old] = nil

  self.settings.options.args[old] = nil
  frames[new] = frames[old]
  frames[old] = nil
  self:BuildTracker(new)
end
function Addon:DeleteTracker(info)
  assertType(info,"table")
  local tracker = info[1]
  self:Delete(self.core, tracker)
  self.settings.options.args[tracker] = nil
end

function Addon:AddMeter(info, name)
  assertType(info,"table", name,"string")
  local trackerName = info[1]
  local tracker = self.core[trackerName]

  if #info < 4 then
    local kind = tracker.add and tracker.add.kind or "spell"
    tracker[kind] = tracker[kind] or {}
    local meter = self:Make(tracker[kind], "meter", name)
    if not meter then return end
    meter.kind = kind
    self:BuildMeter(trackerName, kind, name, meter)
  else
    local meter = tracker[info[2]][info[3]]
    self:AddSubMeter(trackerName, info[2], info[3], meter, meter.kind, name)
  end
end
function Addon:DeleteMeter(info)
  assertType(info,"table")
  info = K(info)
  self:ConfDelete(info:r(1))
  self.settings.crawler:Set(K{ info[1], "args", info[2], "args", info[3] }, nil)
  self:Assoc(self.core[info[1]])
end

function Addon:AddSubMeter(trackerName, kind, meterName, meter, subKind, subName)
  subKind = subKind or "spell"
  assertType(trackerName,"string", kind,"string", meterName,"string", meter,"table", subKind,"string", subName,"string")
  if not self.settings.kinds[subKind] then return end
  meter[subKind] = meter[subKind] or {}
  local subMeter = self:Make(meter[subKind], "subMeter", subName)
  if not subMeter then return end
  meter.unmodified = nil
  local image
  if subKind == "aura" then
    local name,_,image,_,_,_,_,_,_,id = UnitAura("player", subName)
    if subName == name then
      self:Make(subMeter, "alias", id)
    elseif subName == id then
      self:Make(subMeter, "alias", name)
    end
  elseif subKind == "spell" then
    local spellInfo = { GetSpellInfo(subName) }
    image = spellInfo[3]
    local id = spellInfo[7]
    if id and tostring(subName) ~= tostring(id) then
      self:Make(subMeter, "alias", id, true)
    end
  end
  self:BuildMeter(trackerName, kind, meterName, meter)
end
function Addon:DeleteSubMeter(info)
  assertType(info,"table")
  local packedKey = info[#info]
  local packedI = packedKey:find("-")
  local subKind = packedKey:sub(1, packedI - 1)
  local name = packedKey:sub(packedI + 1)
  local trackerName, kind, meterName = unpack(info)
  local meter = self:GetParent(info)
  meter.unmodified = nil
  self:Delete(meter[subKind], name)
  self:BuildMeter(trackerName, kind, meterName, meter)
end

function Addon:AddAlso(info, val)
  assertType(info,"table", val,{"string","number"})
  local trackerName, kind, meterName = unpack(info)
  local meter = self:GetParent(info)
  local also = self:Make(meter, "also", val)
  if not also then return end
  meter.unmodified = nil
  also.image = select(3,GetSpellInfo(val))
  also.color = meter.newColor
  meter.newColor = meter.color
  self:BuildMeter(trackerName, kind, meterName, meter)
end
function Addon:DeleteAlso(info, name)
  assertType(info,"table")
  local meter = self:Get(K(info):r(2))
  meter.unmodified = nil
  self:Delete(meter, name)
  local trackerName, kind, meterName = unpack(info)
  self:BuildMeter(trackerName, kind, meterName, meter)
end
function Addon:TrackerOverride(info, value)
  self:ConfSet(info, value)
  local trackerName = info[1]
  local overrideName = info[#info]

  for kind, name, meter in self:GetAll(self.core[trackerName], "meter", 3) do
    meter[overrideName] = value
  end
end

local function alsoColor(info)
  return K(info):r(1):a(info[#info]:sub(0, -#"color"-1), "color")
end
function Addon:ConfGetAlsoColor(info)
  return self:ConfGet(alsoColor(info))
end
function Addon:ConfSetAlsoColor(info, r, g, b, a)
  return self:ConfSet(alsoColor(info), r, g, b, a)
end

function Addon:SetIcon(meter, image)
  meter.image = image
  return image
end

----------
-- Frames
----------

function Addon:Rebuild(info, r, g, b, a)
  if info then self:ConfSet(info, r, g, b, a) end
  self:BuildTrackerFrame(info[1])
end

function Addon:BuildTracker(trackerName)
  assertType(trackerName,"string")
  local tracker = self.core[trackerName]
  self.settings:BuildTrackerSettings(trackerName, tracker)
  self:BuildTrackerFrame(trackerName, tracker)
  for kind, meterName, meter in self:GetAll(tracker, "meter", 3) do
    self:BuildMeter(trackerName, kind, meterName, meter)
  end
end

function Addon:BuildTrackerFrame(name, tracker)
  if not tracker then tracker = self.core[name] end
  assertType(name,"string", tracker,"table")
  self:AutoDefault(tracker)
  local frame = frames[name]
  if not frame then
    frame = CreateFrame("Frame", self.shortName .. name, UIParent)
    frame.conf = tracker
    frame.tex = frame:CreateTexture(nil,"BACKGROUND")
    frame.tex:SetAllPoints(frame)
    frame.tex:SetColorTexture(0, 1, 0, 0.5)

    frames[name] = frame
    if self.lib.masque then
      frame.msq = self.lib.masque:Group(self.name,name)
    end
  end
  self:Draggable(frame, tracker)
  frame.count = 0
  frame:ClearAllPoints()
  local size = tracker.iconSize
  frame:SetSize(size, size)
  self:Clear(name)
  D_updateFrame(frame, tracker)
  frame:Show()
  return frame
end

function Addon:BuildMeter(trackerName, kind, meterName, meter)
  assertType(trackerName,"string", kind,"string", meterName,"string", meter,"table")
  self:Assoc(self.core[trackerName])
  self.settings:BuildMeterSettings(trackerName, kind, meterName, meter)
end

function Addon:InitButton(button, frame, tracker)
  button:EnableMouse(false)
  button:Disable()
  button.icon = button:CreateTexture()
  --button.icon:SetTexture(missingIcon)
  button.icon:SetAllPoints(button)
  button.text = button:CreateFontString(button:GetName() .. "-Amount")
  button.text:SetShadowOffset(1,-1)
  local animations = button:CreateAnimationGroup()
  local fade = animations:CreateAnimation("Alpha")
  fade:SetStartDelay(tracker.duration)
  fade:SetDuration(fadeDuration)
  fade:SetFromAlpha(1)
  fade:SetToAlpha(0)
  button.fade = fade
  animations:SetScript("OnFinished", function()
    button:Hide()
    self:Queue("Sort", frame)
  end)
  button.animation = animations
end

function Addon:GetFreeButton(trackerName, tracker, kind)
  assertType(trackerName,"string", tracker,"table", kind,"string")
  local frame = frames[trackerName]
  local button
  for _,b in ipairs{ frame:GetChildren() } do
    if not b:IsShown() then
      button = b
      break
    end
  end
  if not button then
    frame.count = frame.count + 1
    button = CreateFrame("Button", frame:GetName() .. "-" .. frame.count, frame)
    self:InitButton(button, frame, tracker)
    if frame.msq then frame.msq:AddButton(button, { Icon = button.icon }) end
  end

  local font = self.lib.media:Fetch(self.lib.media.MediaType.FONT, tracker.font)
  local text = button.text
  text:SetFont(font, tracker.fontSize)
  if kind == "aura" then
    text:SetShadowColor(colUnpack(tracker.specialShadow))
  else
    text:SetShadowColor(colUnpack(tracker.shadow))
  end
  local size = tracker.iconSize
  button:SetSize(size, size)
  text:SetPoint("LEFT", size + 5, 1)
  button.kind = kind
  button:Show()
  return button
end

function Addon:Clear(trackerName)
  assertType(trackerName,"string")
  local children = { frames[trackerName]:GetChildren() }
  if #children == 0 then return end
  for _,child in ipairs(children) do
    self:RemoveAlert(trackerName, child)
  end
end

function Addon:RemoveAlert(trackerName, button)
  assertType(trackerName,"string", button,"table")
  local frame = frames[trackerName]
  button:Hide()
  button:ClearAllPoints()

  --button.icon:SetTexture(missingIcon)
  button.image = nil
  button.kind = nil
  button.meter = nil
  button.text:SetText()
  button.text:SetTextColor(1,1,1,1)
end

function Addon:Sort(frame)
  self:Output("Sorting")
  local children = { frame:GetChildren() }
  if #children == 0 then
    unique:Set{}
    return
  end
  local visible = {}
  for _,child in ipairs(children) do
    if child:IsShown() then
      tinsert(visible, child)
    end
  end
  if #visible == 0 then return end
  nilSort(visible, "time")
  visible[1]:ClearAllPoints()
  visible[1]:SetPoint("TOPLEFT", frame)
  for i=2,#visible do
    visible[i]:ClearAllPoints()
    visible[i]:SetPoint("TOPLEFT", visible[i-1], "BOTTOMLEFT", 0, -spacing)
  end

  if frame.msq then frame.msq:ReSkin() end
end

function Addon:BuildAlert(trackerName, meter, match, meterName, id, alertName, amount, crit, over, time, delay)
  self:Output("Building",alertName,"->",amount,"/",crit,"/",over)
  local kind = meter.kind or "spell"
  local tracker = self.core[trackerName]
  local displayOver = tracker.over
  if displayOver ~= "include" and displayOver ~= "additional" and amount == 0 then return end
  local frame = frames[trackerName]
  local alert
  for _,child in pairs{ frame:GetChildren() } do
    if child.kind == kind and child.alertName == alertName and child:IsShown() then
      alert = child
      break
    end
  end
  if not alert then
    alert = self:GetFreeButton(trackerName, tracker, kind)
    local image
    if meter.override then
      image = meter.image
    elseif kind == "aura" then
      image = select(3, UnitAura("player", match)) or select(3, UnitAura("player", meterName))
    else
      image = select(3, GetSpellInfo(match)) or select(3, GetSpellInfo(meterName))
    end
    if not image and kind == "spell" then image = select(3, GetSpellInfo(id)) end
    local oldImage = meter.image
    if image and not oldImage then meter.image = image end
    if not image and oldImage then image = oldImage end
    local color = meter[match] and meter[match].color
    if not color then color = meter.color end
    if color then alert.text:SetTextColor(colUnpack(color)) end
    alert.icon:SetTexture(image)
    alert.displayCrit = meter.markCrits
    alert.amount = 0
    alert.over = 0
    alert.kind = kind
    alert.alertName = alertName
    alert.time = time
    alert.count = 0
    self:Queue("Sort", frame)
  end

  alert.animation:Stop()
  alert.fade:SetStartDelay(tracker.duration + (delay or 0))
  alert.animation:Play()

  alert.count = alert.count + 1
  alert.amount = alert.amount + amount

  local displayAmount = alert.amount

  if over then alert.over = alert.over + over end
  if displayOver == "include" then
    displayAmount = displayAmount + alert.over
  end

  local amountString = roundString(displayAmount)
  alert.crit = alert.displayCrit and (alert.crit or crit)
  if kind == "spell" and alert.crit then
    amountString = amountString .. "*"
  end
  if alert.over >= 1000 and displayOver == "additional" then
    amountString = amountString .. " (" .. roundString(alert.over) .. ")"
  end
  if meterName == "Touch of Karma" then
    amountString = amountString .. "/" .. roundString(UnitHealthMax("player")/2)
  end
  alert.text:SetText(amountString)

  local hiscores = self.core.hiscore
  if not hiscores[meterName] or hiscores[meterName][1] < displayAmount then
    hiscores[meterName] = { displayAmount, alert.icon:GetTexture(), alert, { alert.text:GetTextColor() }}
  end

  return alert
end


-------------------
-- Prebuilt meters
-------------------

function Addon:UpdateSpec()
  if not inWorld then return end
  player.spec = GetSpecialization()
  self.settings:UpdateSpec()
  for meterName in pairs(self.settings.fromSpellbook) do
    self:AddDefault(meterName)
  end
  for meterName in pairs(self.settings.fromOther) do
    self:AddDefault(meterName)
  end
  for meterName in pairs(self.settings.fromPvp) do
    self:AddDefault(meterName)
  end
end
function Addon:UpdateInv()
  if not inWorld then return end
  self.settings:UpdateInv()

  for itemID in pairs(self.settings.fromItem) do
    itemID = tonumber(itemID)
    self:AddDefault(itemID)
  end
end

function Addon:TemplateKey(name)
  assertType(name,{"string","number"})
  local canons = self.lib.templates.canonical
  local cName
  local key
  local num =  tonumber(name)
  if num then
    cName = canons.item[num]
    if cName then return { "item", cName } end
  end
  cName = canons.all[name]
  if cName then return { "all", cName } end
  cName = canons.race[player.race][name]
  if cName then return { "race", player.race, name } end
  cName = canons.class[player.class][player.spec][name]
  if cName then return { "class", player.class, player.spec, name } end
end

function Addon:AddDefault(name)
  assertType(name,{"string","number"})
  local template = self:MeterFromTemplate(self:TemplateKey(name))
  if template.d then self:AddFromTemplate({"Default Damage", name}, false, true) end
  if template.h then self:AddFromTemplate({"Default Healing", name}, false, true) end
end

function Addon:AddFromTemplate(info, allowExcluded, fromDefault)
  assertType(info,"table")
  local trackerName = info[1]
  local name = info[#info]
  local key = self:TemplateKey(name)
  local template = self:MeterFromTemplate(key)
  if not template then return end
  if template.name and not template.name:find("%%") then name = template.name end
  local kind = template.kind or "spell"
  local tracker = self.core[trackerName]
  tracker[kind] = tracker[kind] or {}
  local trackerWithKind = tracker[kind]
  if allowExcluded == false and self:Excluded(trackerWithKind, name) then
    return
  end
  local meter = trackerWithKind[name]
  if meter and meter.kind and not meter.unmodified then return end
  meter = self:Make(trackerWithKind, "meter", name, true)
  if meter then
    meter.kind = kind
    self:DeployTemplate(meter, template)
    if fromDefault then meter.fromDefault = true end
    template._default = key
    meter.unmodified = true
    self:BuildMeter(trackerName, kind, name, meter)
  end
end

function Addon:DeployTemplate(meter, template)
  assertType(meter,"table", template,"table")
  D.setIndex(meter, template)
  for alsoName,color in pairs(template.also or {}) do
    if not self:Excluded(meter, alsoName) and not self:Overriden(meter, alsoName) then
      local also = self:Make(meter, "also", alsoName)
      also.color = D.hexToCol(color)
    end
  end
  for kind, i, subMeterName in D.npairs(template.meters, 3) do
    meter[kind] = meter[kind] or {}
    local meterWithKind = meter[kind] or {}
    if not self:Excluded(meterWithKind, subMeterName) and not self:Overriden(meterWithKind, subMeterName) then
      self:Make(meterWithKind, "subMeter", subMeterName)
    end end
end

function Addon:MeterFromTemplate(key)
  if not key then return end
  assertType(key,"table")
  local cached = templateCache:Get(key)
  if cached then return cached end
  local template = D.crawl(self.lib.templates.aliases, key, true)
  if not template then return end
  local name = key[#key]
  local meter = {}
  for k,v in pairs(template) do meter[k] = v end

  if type(name) == "number" then
    meter.image = GetItemIcon(name)
    if meter.override ~= false then meter.override = true end
  elseif meter.override and meter.id then
    meter.image = select(3, GetSpellInfo(meter.id))
  end
  if meter.color then meter.color = D.hexToCol(meter.color) end
  for also,color in pairs(template.also or {}) do
    local icon = select(3, GetSpellInfo(also))
    if icon then
      meter.image = meter.image or icon
      meter.color = meter.color or D.hexToCol(color)
    end
  end
  if not meter.image then
    meter.image = select(3, GetSpellInfo(meter.name)) or (select(3,GetSpellInfo(name))) or (select(3,GetSpellInfo(template.id)))
  end
  meter.newColor = meter.color
  meter.kind = template.kind or "spell"
  setIndex(meter, self.defaults.meter, true)
  meter._default = key
  templateCache:Set(key, meter)
  return meter
end


----------
-- Events
----------

function Addon:PLAYER_ENTERING_WORLD()
  player.id = UnitGUID("player")
  player.name = UnitName("player")
  player.race = select(2, UnitRace("player"))
  player.class = select(2, UnitClass("player"))
  player.spec = GetSpecialization()
  if not inWorld then
    local settingsCrawler = self.settings.crawler
    for tracker, kind, meterName, meter in self:GetAll(self.core, "meter", 4) do
      local unmodified = meter.unmodified
      local key = K{ tracker, kind, meterName }
      local delete = false
      local templateKey = meter._default
      local template = self:MeterFromTemplate(templateKey)
      if templateKey and not template then
        delete = true
        templateKey = nil
      end

      if not templateKey and (unmodified or meter.fromDefault) then
        templateKey = self:TemplateKey(meterName)
        template = self:MeterFromTemplate(templateKey)
        if template then
          meter._default = templateKey
        else
          delete = true
        end
      end
      if templateKey then
        delete = true
        if template then
          self:DeployTemplate(meter, template)

          local templateKind = template.kind or "spell"
          if templateKind ~= kind and unmodified then
            delete = true
            self.core[tracker][templateKind][meterName] = meter
            settingsCrawler:Set({ tracker, "args", templateKind, "args", meterName }, nil)
            self:BuildMeter(tracker, templateKind, meterName, meter)
          elseif meter.fromDefault and unmodified then
            delete = true
          else
            delete = false
          end
        end
      end
      if delete then
        settingsCrawler:Set({ tracker, "args", kind, "args", meterName }, nil)
        self.core[tracker][kind][meterName].type = nil
      end
      meter.newColor = meter.color
    end
  end

  inWorld = true
  self:UpdateSpec()
  self:UpdateInv()
end

local DELIM = "`"

function Addon:UpdateLog(logged, amount, overflow, crit, time)
  logged.amount = (logged.amount or 0) + (amount or 0) - (overflow or 0)
  logged.overflow = (logged.overflow or 0) + (overflow or 0)
  logged.crit = logged.crit or crit
  logged.time = logged.time or time
end

function Addon:Match(match, trackerName, time, seconds, destGUID, spellID, spellName, amount, crit, overflow)
  if not match then return end
  for meterName, meter in pairs(match) do
    local finalSuff = seconds
    local alertName = meter.name or meterName
    local suff = ""
    local spell, aura, recent
    spell = unique:Get{ "spell", meterName }
    if destGUID then aura = unique:Get{ "aura", meterName, destGUID } end
    recent = unique:Get{ "recent", meterName }
    self:Output(meterName,"-",aura,"/",spell,"/",recent and recent[2])
    if aura then
      suff = suff .. aura
      finalSuff = ""
    elseif spell then
      suff = suff .. spell
      finalSuff = ""
    elseif recent and seconds - recent[2] < 2 then
      suff = suff .. recent[1]
      finalSuff = ""
    elseif recent then
      recent = { seconds }
    end
    if not recent then recent = { seconds } end
    recent[2] = seconds

    unique:Set({ "recent", meterName }, recent)
    local display = meter.display
    if (amount or 0) + (overflow or 0) > 0 then
      if display ~= "aggregate" then
        alertName = alertName .. suff .. finalSuff
        finalSuff = ""
      end
      if display == "individual" then
        alertName = alertName .. finalSuff .. unique:Incr{ "time", time }
        finalSuff = ""
      end

      self:BuildAlert(trackerName, meter, spellName, meterName, spellID, alertName,
        amount, crit, overflow, time)
    end
  end
end

function Addon:COMBAT_LOG_EVENT_UNFILTERED(
event,time,suffix,hide,
sourceGUID,sourceName,sourceFlags,sourceRaidFlags,
destGUID,destName,destFlags,destRaidFlags,
spellID,spellName,school,amount, overflow, school, resisted, blocked, absorbed, crit
)
  local kind
  if suffix == "SPELL_CAST_SUCCESS" then
    kind = "spell"
  elseif suffix == "SWING_DAMAGE" then
    amount = spellID
    spellID = nil
    spellName = nil
    kind = "damage"
  elseif suffix == "SPELL_DAMAGE" or suffix == "SPELL_BUILDING_DAMAGE"
      or suffix == "SPELL_PERIODIC_DAMAGE" or suffix == "RANGE_DAMAGE" then
    kind = "damage"
  elseif suffix == "SPELL_HEAL" or suffix == "SPELL_BUILDING_HEAL"
      or suffix == "SPELL_PERIODIC_HEAL" or suffix == "RANGE_HEAL" then
    kind = "healing"
    absorbed = school
    crit = resisted
  elseif suffix == "SPELL_AURA_APPLIED" or suffix == "SPELL_AURA_REFRESH" or suffix == "SPELL_AURA_REMOVED" then
    kind = "aura"
  else return end
  local unitMatch = sourceGUID == player.id
  local petMatch = not unitMatch and (
  CombatLog_Object_IsA(sourceFlags, COMBATLOG_FILTER_MY_PET) or CombatLog_Object_IsA(sourceFlags, COMBATLOG_FILTER_MINE)
  )

  local destMatch = destGUID == player.id or (
  CombatLog_Object_IsA(destFlags, COMBATLOG_FILTER_MY_PET) or CombatLog_Object_IsA(destFlags, COMBATLOG_FILTER_MINE)
  )
  if destMatch and kind == "damage" then return end

  if not unitMatch and not petMatch then return end

  local seconds = tostring(floor(time))

  if type(spellName) == "string" then
    spellName = spellName:gsub(" Off-Hand",""):gsub(" Overload","")
  end

  if unitMatch then
    if kind == "aura" then
      if destGUID and destGUID ~= player.id then
        self:Output("Unique aura",spellName,destGUID,seconds)
        unique:Set({ "aura", spellName or sourceName, destGUID }, unique:Get{ "dest", spellName, destGUID } or seconds)
      end
      return
    elseif kind == "spell" then
      self:Output("Unique spell",spellName,seconds)
      unique:Set({ "spell", spellName }, seconds)
      if destGUID and destGUID ~= player.id then
        unique:Set({ "dest", spellName, destGUID }, seconds)
      end
      return
    end
  elseif kind =="aura" or kind == "spell" then
    return
  end

  if sourceName == player.name then sourceName = "%n" end
  local key = DELIM:join(
    tostring(kind),
    tostring(time),
    tostring(seconds),
    tostring(sourceName),
    tostring(petMatch),
    tostring(destGUID),
    tostring(spellID),
    tostring(spellName))
  self:Output("Logging",key)
  logKeys[key] = { kind, time, seconds, sourceName, petMatch, destGUID, spellID, spellName }
  local logged = log[key]
  if not logged then
    logged = {}
    log[key] = logged
  end
  self:UpdateLog(logged, amount, overflow, crit, time)
  if not unitMatch then
    self:UpdateLog(logTotals:GetOrMake{ kind, sourceName }, amount, overflow, crit, time)
  end
  self:UpdateLog(logTotals:GetOrMake{ kind, "_total" }, amount, overflow, crit, time)
  self:Queue("ResolveLog")
end

function Addon:ResolveLog()
  self:Output("Processing")
  local lastLog = log
  local lastKeys = logKeys
  local lastTotals = logTotals.core
  log = {}
  logKeys = {}
  logTotals.core = {}

  local core = self.core
  for trackerName,trackerFrame in pairs(frames) do
    local tracker = core[trackerName]
    if tracker and tracker.enabled and lastTotals[tracker.trackerType] then
      local totals = lastTotals[tracker.trackerType]
      local total = totals._total
      local t_amount, t_crit, t_overflow = total.amount, total.crit, total.overflow
      local t_time = total.time - 9000
      for auraName, meterName, meter in self:GetAll(tracker.assoc.core.aura, "meter", 3) do
        meterName = meter.name or meterName
        local name,_,icon,_,_,_,expires,_,_,_,id = UnitAura("player", auraName)
        if name then
          self:BuildAlert(trackerName, meter, auraName, meterName, id, meterName,
            t_amount, t_crit, t_overflow, t_time, expires - GetTime())
        end
      end
      for petName, meterName, meter in self:GetAll(tracker.assoc.core.pet, "meter", 3) do
        meterName = meter.name or meterName
        local petTotal = totals[petName]
        if petTotal then
          local pt_amount, pt_crit, pt_overflow = petTotal.amount, petTotal.crit, petTotal.overflow
          local pt_time = petTotal.time - 8000
          self:BuildAlert(trackerName, meter, petName, meterName, nil, meterName,
            pt_amount, pt_crit, pt_overflow, pt_time)
        end
      end
    end
  end

  for key,logged in pairs(lastLog or {}) do
    local kind, time, seconds, sourceName, petMatch, destGUID, spellID, spellName = unpack(lastKeys[key])
    local amount = logged.amount
    local overflow = logged.overflow
    local crit = logged.crit
    for trackerName,trackerFrame in pairs(frames) do
      local tracker = core[trackerName]
      if tracker and tracker.enabled and (not kind or kind == tracker.trackerType) and tracker.assoc and tracker.assoc.core.spell then
        self:Match(tracker.assoc.core.spell[spellName],
          trackerName, time, seconds, destGUID, spellID, spellName,
          amount, crit, overflow)
        self:Match(tracker.assoc.core.spell[spellID],
          trackerName, time, seconds, destGUID, spellID, spellName,
          amount, crit, overflow)
      end
    end
  end
end